// AppController.m -- implementation of the stackinator application controller.
//		      The prefs and the stack contents are put into the
//		      userprefs out of sheer slot

#import "AppController.h"
#import "KeyCombo.h"
#import "HotKeyCenter.h"
#import "KeyComboPanel.h"
#import <Carbon/Carbon.h>


// forward declarations of methods
// This keeps them out of the header file

@interface AppController(Private)

- (void) load: (id) sender;
- (void) messWithHotKey: (BOOL) onOff;

@end // Private method(s)



@implementation AppController


// do random one-time initialization

- (id) init
{
    if (self = [super init]) {

	// make our model object
	items = [[NSMutableArray alloc] init];

	// defaulth ot key (but turned off) is cmd-shift-option-return
	hotkeyCombo = [KeyCombo keyComboWithKeyCode: 36	// return
				andModifiers: cmdKey | shiftKey | optionKey];
	[hotkeyCombo retain];

	// read user data from the preferences
	[self load: self];
    }

    return (self);

} // init


// clean up our mess.  Will probably never get called since 
// the app controller is instantiated in the main nib file,
// but it's good to do.

- (void) dealloc
{
    [items release];
    [hotkeyCombo release];

    [super dealloc];

} // dealloc



// the nib has been loaded and connections set up

- (void) awakeFromNib
{
    // only do this work once.  awakeFromNib will be called again
    // when the pref window is shown.  We definitely don't want to
    // do this initialization work twice

    if (!alreadyAwake) {

	// tweak window settings.  The 'utility window' NSPanel
	// is set to float over other windows, and hide on deactivate.
	// Stackinator only wants a utility window panel to get the
	// skinny title bar
	[window setLevel: NSNormalWindowLevel];
	[window setHidesOnDeactivate: NO];

	// set up the 'close' cell - an X button like in mail.app or
	// itoots that will end up deleting a row.  This is just an
	// NSButtonCell with a pair of images, one for 'resting' state,
	// and one for 'mouse tracking' state.

	NSTableColumn *column;
	column = [stackTableview tableColumnWithIdentifier: @"closer"];

	NSBundle *bundle;
	bundle = [NSBundle bundleForClass: [self class]];

	NSString *path;

	path = [bundle pathForResource: @"x-mask" ofType: @"tiff"];
	NSImage *xImage;
	xImage = [[NSImage alloc] initWithContentsOfFile: path];

	path = [bundle pathForResource: @"x-highlight" ofType: @"tiff"];
	NSImage *xhighlightImage;
	xhighlightImage = [[NSImage alloc] initWithContentsOfFile: path];

	// set up the button cell
	NSButtonCell *cell;
	cell = [[NSButtonCell alloc] init];

	[cell setButtonType: NSToggleButton];
	[cell setImage: xImage];
	[cell setAlternateImage: xhighlightImage];
	[cell setBordered: NO];

	// what to do when it gets clicked
	[cell setTarget: self];
	[cell setAction: @selector(deleteRow:)];

	// and tell the column (right-hand-most column) to use it
	[column setDataCell: cell];

	// we've done our one-time initialization.  don't do it again.
	alreadyAwake = YES;
    }

} // awakeFromNib



// add a new item to the top of the stack.  This is the target of
// both a pushbutton and the text field

- (IBAction) pushNewItem: (id) sender
{
    NSString *newItem;
    newItem = [[[newItemTextfield stringValue] copy] autorelease];

    // only insert non-empty items
    if ([newItem length] > 0) {
	[items insertObject: newItem  atIndex: 0];
	
	[stackTableview reloadData];

	// clear out the textfield
	[newItemTextfield setStringValue: @""];

	// and save the list for paranoia's sake
	[self save: self];
    }

} // pushNewItem



// remove a row.  This is called from the NSButtonCell in the
// tableview

- (void) deleteRow: (id) sender
{
    int row;
    row = [stackTableview selectedRow];

    if (row >= 0) {
	[items removeObjectAtIndex: row];
	[stackTableview reloadData];
	[self save: self];
    }

} // deleteRow



// store our fiddly bits to the user prefs

- (void) save: (id) sender
{
    NSUserDefaults *defaults;
    defaults = [NSUserDefaults standardUserDefaults];

    // save our model
    [defaults setObject: items  forKey: @"items"];

    // hotkey stuff
    [defaults setInteger: [hotkeyCombo keyCode]
	      forKey: @"hotkeyCode"];
    [defaults setInteger: [hotkeyCombo modifiers]
	      forKey: @"hotkeyModifiers"];

    // whether we stored the hotkey.  This lets us know
    // the difference between "0/0 for no hotkey" vs
    // "0/0 for being freshly initialized"
    [defaults setBool: YES  forKey: @"keyComboStored"];

    // the checkbox values
    [defaults setBool: hotkeyActive  forKey: @"hotkeyActive"];
    [defaults setBool: floatOverAllWindows  forKey: @"floatOverAllWindows"];

    // and force them to be written to the disk (hopefully)
    [defaults synchronize];

} // save



// restore our random information from the user prefs

- (void) load: (id) sender
{
    NSUserDefaults *defaults;
    defaults = [NSUserDefaults standardUserDefaults];

    NSArray *array;
    array = [defaults objectForKey: @"items"];

    // replace our array's contents with the loaded array
    [items setArray: array];

    int comboStored;
    comboStored = [defaults boolForKey: @"keyComboStored"];

    if (comboStored) {
	
	// get the individual parts of the keycode, then
	// piece it together
	short keycode;
	short modifiers;
	keycode = [defaults integerForKey: @"hotkeyCode"];
	modifiers = [defaults integerForKey: @"hotkeyModifiers"];

	// replace the current keycode with a new one
	[hotkeyCombo release];
	hotkeyCombo = [KeyCombo keyComboWithKeyCode: keycode
				andModifiers: modifiers];
	[hotkeyCombo retain];

	// restore preference, and actually make it affect
	// program behavior
	floatOverAllWindows = [defaults boolForKey: @"floatOverAllWindows"];

	// install the hotkey
	hotkeyActive = [defaults boolForKey: @"keyComboStored"];
	[self messWithHotKey: hotkeyActive];
    }

} // load



// --------------------------------------------------
// tableview datasource

// how many rows in our table?

- (int) numberOfRowsInTableView: (NSTableView *) tableView
{
    int count;
    count = [items count];

    return (count);

} // numberOfRowsInTableView



// what is the name of that doggie in the window?

- (id) tableView: (NSTableView *) tableView  objectValueForTableColumn: (NSTableColumn *) tableColumn
	     row: (int) row
{
    NSString *value;
    value = [items objectAtIndex: row];

    return (value);

} // objectValueForTableColumn


// a new one provided by BWDeletableTableView - the user pressed 
// the delete key while the tableview had focus.  Remove the selected
// row.

- (void) tableView: (NSTableView *) tableView
   shouldDeleteRow: (int) row
{
    [items removeObjectAtIndex: row];
    [stackTableview reloadData];

    [self save: self];

} // shouldDeleteRow



// --------------------------------------------------
// Hotkey stuff
// courtesy of Quentin Carnicelli
// http://www.rogueamoeba.com/~qdc/pub/


// depending on an on/off state, activate or deactivate the
// hot key which brings the stackinator window forward

- (void) messWithHotKey: (BOOL) onOff
{
    HotKeyCenter *sharedCenter;
    sharedCenter = [HotKeyCenter sharedCenter];

    if (onOff) {
	// add the key
	BOOL good;
	good = [sharedCenter addHotKey: @"Stackinator Summons HotKey"
			     combo: hotkeyCombo
			     target: self
			     action: @selector(summonHotkeyDown:)];
	if (!good) {
	    NSLog (@"could not register hotkey");
	}
	hotkeyActive = YES;
	
    } else {
	// remove the key
	[sharedCenter removeHotKey: @"Stackinator Summons HotKey"];
	hotkeyActive = NO;
    }

} // messWithHotKey



// this is invoked when our global hot-key is pressed.  Bring the
// stackinator window to the front and give focus tot he text field

- (void) summonHotkeyDown: (id) sender
{
    // !!! it'd be nice that if we were already active, we could
    // !!! move ourselves to the background.  As far as I could find,
    // !!! this isn't possible in a clean way.

    [window makeKeyAndOrderFront: self];

    // sometimes the window will come front but the app not activate.
    // this forces us to become the active application
    [NSApp activateIgnoringOtherApps: YES];

} // summonHotkeyDown



// --------------------------------------------------
// Preference window stuff


// UI action to toggle the 'float over everyone else' setting

- (IBAction) changeFloatAboveAllWindows: (id) sender
{
    if ([sender state] == NSOnState) {
	// NSDockWindowLevel is officially deprecated, but there's
	// not a recommended alternative.
	// here we want to float just below the dock, so the user can
	// get to the dock if they want to

	[window setLevel: NSDockWindowLevel - 1];

	floatOverAllWindows = YES;

    } else {
	[window setLevel: NSNormalWindowLevel];
	floatOverAllWindows = NO;
    }

} // changeFloatAboveAllWindows


// UI action to toggle the 'hot key brings us forward' setting

- (IBAction) changeActivateHotKey: (id) sender
{
    [self messWithHotKey: [sender state] == NSOnState];

} // changeActivateHotKey



// UI action to chagne the hot key.  It shows the Key Combo Panel
// modallly.

- (IBAction) changeHotKey: (id) sender
{
    // get the panel
    KeyComboPanel *panel;
    panel = [KeyComboPanel sharedPanel];

    // let it know what the current setting is
    [panel setKeyCombo: hotkeyCombo];

    // interact with the user
    int result;
    result = [panel runModal];

    // get any new value
    if (result == NSOKButton) {
	[hotkeyCombo release];
	hotkeyCombo = [[panel keyCombo] retain];

	[hotkeyDescription setStringValue: [hotkeyCombo userDisplayRep]];

	if (hotkeyActive) {
	    [self messWithHotKey: YES];
	}

	// store the new key combination in the preferences
	[self save: self];
    }

} // changeHotKey



// UIAction to show the preference window

- (IBAction) showPreferences: (id) sender
{
    // update the UI elements
    [hotkeyDescription setStringValue: [hotkeyCombo userDisplayRep]];

    if (floatOverAllWindows) {
	[floatCheckbox setState: NSOnState];
    }
    if (hotkeyActive) {
	[hotkeyCheckbox setState: NSOnState];
    }

    // and bring it forward
    [prefsWindow makeKeyAndOrderFront: self];

} // showPreferences



// --------------------------------------------------
// Application delegate and window handling

// This is a gross hack.  The behavior it fixes is this:
//   * stackinator panel window and the preference window are open
//   * user closes the preference window. (or the about box)
//   * NSApp asks 'should terminate after last window closed', even
//     though the panel is still open (I guess it doesn't count as a real
//     window)

// so, to make this not happen, if the window being closed is the
// stack window, set a flag to tell the app to terminate after last
// window close.  If the panel window closes, that's OK.  We do want
// to quit in that case.

static BOOL allowTerminate;

- (BOOL) windowShouldClose: (id) sender
{
    if (sender == window) {
	allowTerminate = YES;
    }

    return (YES);

} // windowSouldClose



// quit the app if there are no more windows showing
// there is some controversy to this general behavior:
//   http://cocoa.mamasam.com/COCOADEV/2002/11/1/50036.php
// my opinion is that Stackinator is a small utility that has no
// purpose after the window closes, plus it's fast to start, that this
// is the correct bheavior

- (BOOL) applicationShouldTerminateAfterLastWindowClosed: (id) sender
{
    return (allowTerminate);

} // applicationShouldTerminateAfterLastWindowClosed



// app is going down.  save our stuff.

- (NSApplicationTerminateReply) applicationShouldTerminate: (NSApplication *) app
{
    [self save: self];

    return (NSTerminateNow);

} // applicationShouldTerminate


@end // AppController
